import pathlib

from prestoplot import storages, story

PATH = pathlib.Path(__file__).parent

DATA = PATH / "data"


def test_render_story():
    storage = storages.FileStorage(DATA)
    result = story.render_story(storage, "characters", seed="testing")
    expected = (
        "Our hero, Noah, is a flashing-eyed, erring explorer, "
        "subjected to adverse conditions. He carries a whip."
    )
    assert result == expected


def test_render_story_jinja2():
    storage = storages.FileStorage(DATA)
    result = story.render_story(storage, "characters_jinja", seed="testing")
    expected = (
        "Our hero, Noah, is a flashing-eyed, erring explorer, "
        "subjected to adverse conditions. He carries a whip."
    )
    assert result == expected



import logging
import math
import random

from . import contexts, markov
from .texts import is_text


class Database:
    def __init__(self, factory, grammar_path, context):
        self.factory = factory
        self.context = context
        self.grammar_path = grammar_path
        self.cache = {}

    def __getattr__(self, attr):
        if attr not in self.cache:
            try:
                result = self.__dict__[attr]
            except KeyError:
                if hasattr(str, attr):
                    result = getattr(str(self), attr)
                else:
                    seed = f'{self.context["seed"]}-{attr}'
                    with contexts.update_context(self.context, key=attr, seed=seed):
                        result = self.factory(self.context)
            self.cache[attr] = result
        return self.cache[attr]

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except TypeError:
            logging.exception(f"No key {key!r} in {self!r}")
            raise

    def __str__(self):
        return str(self.factory(self.context))


class Databag(dict):
    def __init__(self, grammar_path, context, *args, **kwargs):
        self.context = context
        self.grammar_path = grammar_path
        super().__init__(*args, **kwargs)

    def __getattr__(self, attr):
        return self[attr]

    def __getitem__(self, key):
        try:
            value = super().__getitem__(str(key))
        except KeyError:
            logging.error(f'No key {key!r} in {", ".join(self.keys())}')
            raise
        if is_text(value):
            value = value.render(self.context)

        return value


class Datalist(list):
    def __init__(self, grammar_path, context, *args, **kwargs):
        self.context = context
        self.grammar_path = grammar_path
        super().__init__(*args, **kwargs)

    def __getitem__(self, idx):
        try:
            value = super().__getitem__(idx)
        except KeyError:
            logging.error(f'No index {idx!r} in {", ".join(self)}')
            raise
        if is_text(value):
            value = value.render(self.context)

        return value


def choose(items):
    def chooser(context):
        rng = random.Random(context["seed"])
        result = rng.choice(items)
        if is_text(result):
            result = result.render(context)
        return result

    return chooser


def pick(items):
    def picker(context):
        rng = random.Random(context["seed"])
        if len(items) > 1:
            idx = rng.randint(0, len(items) - 1)
        else:
            idx = 0
        result = items.pop(idx)
        if is_text(result):
            result = result.render(context)
        return result

    return picker


def markovify(items):
    def generator(context):
        gen = markov.NameGenerator(items, chainlen=context.get("markov_chainlen", 2))

        return gen.get_random_name(start=context.get("start_markov", ""))

    return generator


def pareto_int(rng, shape=1):
    return math.floor(rng.paretovariate(shape))



import atexit
import logging
import pathlib
import sys
import textwrap

import click

from . import storages, story


@click.group()
@click.option("--debug", is_flag=True)
@click.option("--pdb", is_flag=True)
def main(debug, pdb):
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)
    if pdb:

        @atexit.register
        def debug_on_exit():
            if hasattr(sys, "last_traceback"):
                try:
                    import ipdb as pdb
                except ImportError:
                    import pdb
                pdb.pm()


@main.command()
@click.argument("path")
@click.option(
    "--count", default=1, help="How many times to consult the oracle. Default 1."
)
@click.option(
    "--markov-start",
    default="",
    help="Characters to start any Markov chains with, e.g. 'Ba'. Default ''.",
)
@click.option(
    "--markov-chainlen", default=2, help="Length of Markov chain links. Default 2."
)
@click.option("--wrap", is_flag=True, help="Wrap text?")
@click.option(
    "--wrap-length", default=78, help="Maximum line length when wrapping text."
)
@click.option("--seed", help="Pre-seed the random generator.")
def run(path, count, markov_start, markov_chainlen, wrap, wrap_length, seed=None):
    """Parse and consult a YAML generative grammar oracle file.

    The "oracle" consulted directly must include a `Begin:` stanza.

    Example:

    \b
    $ cat names.yaml
    Begin:
      - "{Name}"
    \b
    Name:
      - George
      - Martha
    \b
    $ presto run names.yaml
    George

    """
    if markov_start and len(markov_start) < markov_chainlen:
        raise click.UsageError(
            f"--markov-start must be at least as long as "
            f"--markov-chainlen, currently {markov_chainlen}."
        )
    kwargs = dict(start_markov=markov_start, markov_chainlen=markov_chainlen)
    if seed is not None:
        kwargs["seed"] = seed

    path = pathlib.Path(path)
    storage = storages.FileStorage(path.parent)

    for n in range(count):
        result = story.render_story(storage, path.stem, **kwargs)
        if not wrap:
            print(result)
        else:
            for line in result.splitlines():
                if line.strip():
                    for new_line in textwrap.wrap(line, wrap_length):
                        print(new_line)
                else:
                    print(line)
        if n + 1 != count:
            print("\n---\n")



import pathlib

import pytest
import yaml

from prestoplot import storages

PATH = pathlib.Path(__file__).parent

DATA = PATH / "data"


@pytest.fixture(
    params=[
        storages.FileStorage,
        storages.CachedFileStorage,
        storages.CompilingFileStorage,
    ]
)
def fs(request):
    return request.param(str(DATA))


def test_file_storage_should_store_path(fs):
    assert isinstance(fs.path, pathlib.Path)
    assert fs.path == DATA


def test_file_storage_should_list_modules(fs):
    result = fs.list_modules()
    assert result == ["characters", "characters_jinja", "names"]


def test_file_storage_resolve_module_should_resolve_valid_names(fs):
    result = fs.resolve_module("names")
    with open(DATA / "names.yaml") as fi:
        expected = yaml.safe_load(fi)
    assert result == expected


def test_file_storage_resolve_module_should_not_resolve_invalid_names(fs):
    with pytest.raises(storages.ModuleNotFoundError):
        fs.resolve_module("foo")



import pathlib

from prestoplot import texts

PATH = pathlib.Path(__file__).parent

DATA = PATH / "data"


def test_text_indeterminate_article_a():
    text = texts.Text("foo", "path", {})
    assert text.a == "a"
    assert text.an == "a"


def test_text_indeterminate_article_an():
    text = texts.Text("ollie", "path", {})
    assert text.a == "an"
    assert text.an == "an"


def test_renderable_text_indeterminate_article_a():
    text = texts.RenderableText("foo", "path", {})
    assert text.a == "a"
    assert text.an == "a"


def test_renderable_text_indeterminate_article_an():
    text = texts.RenderableText("ollie", "path", {})
    assert text.a == "an"
    assert text.an == "an"



import hashlib
import random


def make_seed(rng=random):
    return hashlib.md5(str(rng.random()).encode("utf-8")).hexdigest()


def get_seed(context):
    try:
        seed = context["seed"]
    except KeyError:
        seed = context["seed"] = make_seed()

    key = context.get("key")
    if key is not None:
        seed = f"{seed}-{key}"

    return seed



import logging

import jinja2
from funcy import cached_property, identity, isa

jinja2_env = jinja2.Environment(undefined=jinja2.DebugUndefined)


def render_ftemplate(tmpl, grammar_path, context):
    global_ctx = {**context}
    local_ctx = {"result": tmpl}
    try:
        exec("result = eval(f'''f{result!r}''')", global_ctx, local_ctx)
    except Exception as exc:
        logging.error(f"{exc}--Could not render template in {grammar_path}:")
        logging.error(f"Template:\n{tmpl}")
        raise
    return RenderedStr(local_ctx["result"])


def render_jinja2(tmpl, grammar_path, context):
    try:
        return RenderedStr(jinja2_env.from_string(tmpl).render(context))
    except jinja2.TemplateError as exc:
        logging.error(f"{exc}--Could not render Jinja2 template in {grammar_path}:")
        logging.error(f"Template:\n{tmpl}")
        msg = f"Could not render Jinja2 template ({exc}): "
        if hasattr(exc, "source") and exc.source:
            lineno = exc.lineno
            line = exc.source.splitlines()[lineno - 1]
            logging.error(f"{grammar_path}, line {lineno}")
            logging.error(f"--> {line}")
            msg += f"\n{grammar_path}, line {lineno}"
            msg += f"\n--> {line}"
        else:
            msg += f"\n{grammar_path}"
            msg += f"\n{tmpl}"
        raise ValueError(msg)


class Text:
    vowels = set("aeiou")

    def __init__(self, value, grammar_path, context, transformer=identity):
        self.value = value
        self.context = context
        self.grammar_path = grammar_path
        self.transformer = transformer

    @cached_property
    def an(self):
        if str(self.value)[0].lower() in self.vowels:
            return "an"
        else:
            return "a"

    a = an

    @cached_property
    def An(self):
        return self.an.capitalize()

    A = An

    def __str__(self):
        return self.render()

    def __repr__(self):
        return repr(self.value)

    def __getattr__(self, attr):
        if hasattr(str, attr):
            return getattr(self.value, attr)
        else:
            return super().__getattr__(attr)

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        return str(self) == other

    def render(self, context=None):
        return RenderedStr(self.transformer(self.value))


class RenderableText(Text):
    def __init__(
        self,
        value,
        grammar_path,
        context,
        transformer=identity,
        render_strategy=render_ftemplate,
    ):
        super().__init__(value, grammar_path, context, transformer=transformer)
        self._render = render_strategy

    def __str__(self):
        return self.render(self.context)

    def render(self, context=None):
        return RenderedStr(
            self._render(super().render(), self.grammar_path, context or self.context)
        )


is_text = isa(Text)


class RenderedStr(str):
    vowels = set("aeiou")

    @cached_property
    def an(self):
        if self[0].lower() in self.vowels:
            return "an"
        else:
            return "a"

    a = an

    @cached_property
    def An(self):
        return self.an.capitalize()

    A = An
